%{
MAMSO - Multi-Agents Multi-Strategies Optimiser
 
A tool to easily compare many combinations
of different strategies.

Version
 2021-02-21 Maurice.Clerc@WriteMe.com

The problems are described in getFunc.m
The search space can be discrete (see the 'quantis' vector)


Warning
-------
For the local search the program uses an _embedded_ one
However the one of Matlab seems to be better than the one of Octave.


Know bugs and errors
---------------------
- With Octave, some problems of CEC2011 can not be solved, because
  their definitions use specific Matlab commands.
- Octave, Dice problem, if local search => error: objective function not real
- With Matlab, on some problems, like Pressure Vessel, if local search
=> Matrix is close to singular or badly scaled. Results may be inaccurate.

Examples
--------
-------------------------------------  MPA-like (Marine Predator Algorithm)
(improved, for the original version is translation sensitive, and
also unnecessarily complicated)
sequenceOption=1;
sequenceList=[1 2 3];
outLoop=[ 4 5  13 ];

mamsoloop(100,25,6000,'Dice'); => -1.578969

Simplified + local search after each iteration localFrequency=1)
    outLoop=[ 9 ];
 =>  -1.610825

 Simplified +
   localFrequency=0.75 => -1.600993
   localFrequency=0.5  => -1.611657
   localFrequency=0.25  => -1.599446
             

Just a test:
sequenceList=[ones(1,79) 9 2*ones(1,79) 9 3*ones(1,41) 9 ];
outLoop=[ ];
sequenceOption=0;
 => -1.583334

 ----------------------------------------------------------------- PSO-like
mamsoloop(100,25,6000,'Dice');
sequenceOption=0;
sequenceList=[6];
outLoop=[ 9 ];
localFrequency=1;
=> 1.601585

with localFrequency=0.5 => -1.612625
with localFrequency=1/3 => -1.611786
with localFrequency=0.25 => -1.613224
with localFrequency=0.2 => -1.611678
with localFrequency=0.1 => -1.612570

---
mamsoloop(100,25,6000,'Gear_train');
sequenceOption=0;
sequenceList=[6];
outLoop=[ 9 ];
localFrequency=1;
=> 2.700857e-12

------------------------------------------------------------------ APS-like
mamsoloop(100,25,6000,'Dice');
sequenceList=[7 8 ];
outLoop=[ 9 10 11 12 ];
sequenceOption=0;
-1.527394

sequenceOption=1; =>  -1.581739 
sequenceOption=2; => -1.589128 

Hybrid 
sequenceList=[7 8 14 ];
outLoop=[ 9  ];
sequenceOption=1;
=> -1.576699

---
mamsoloop(100,25,6000,'Gear_train');
sequenceList=[7 8];
outLoop=[ 9 10 11 12];
sequenceOption=2;
=> 9.939876e-11 

outLoop=[9]
sequenceOption=0;
=> 2.700857e-12 

------------------------------------------------------------------ EDA-like
mamsoloop(100,25,6000,'Dice');
sequenceList=[14];
outLoop=[9];
localFrequency=1 =>   -1.607897
localFrequency=0.5 => -1.587635 

outLoop=[ 9 10 11]
localFrequency=1 => -1.604457
localFraquency=0.75 => -1.570185
localFrequency=0.5 =>  -1.596447
localFrequency=0.25 => -1.585518

----
mamsoloop(100,25,6000,'Gear_train');
sequenceOption=0;
sequenceList=[14];
outLoop=[ 9 ];
localFrequency=1;
=> 2.700857e-12

---
mamsoloop(15,25,25000,'CEC2019_nn');
1.000000
9.525336
1.000000
2.989918
1.000000
5.774814
4.789687
3.639723
1.050758
20.06946

-----------------------------  AOA-like (Archimedes Optimization Algorithm)
(seriously improved, for the original version is translation sensitive, and
unnecessarily complicated with too many parameters)

mamsoloop(100,25,6000,'Dice');
sequenceOption=0;
sequenceList=[15];
outLoop=[ 4  ];
=> -1.611894
outLoop=[ 4 9 ]; => -1.612181
---
mamsoloop(100,25,6000,'Gear_train');
outLoop=[ 4  ]; => 2.700857e-12
outLoop=[ 4 9 ]; => 2.700857e-12

---  CEC 2019
mamsoloop(15,25,25000,'CEC2019_nn');
outLoop=[ 4  ];
1413.899957
75.594953
1.002947
4.979836
1.017236
1
10.363001
1.782092
1.044827
1.000303

outLoop=[ 4 9 ]
1
51.608356
1.409135
5.974795
1
1.01647
261.540562
3.053764
1.038137
20.998664



--------------------------------------------------------------------- TO DO
- Automatically select the sets of strategies, depending on the dimension?
Or, at least, define guidelines
- More strategies. GA? ACO? 
- More problems


======================================================== List of strategies
1) (coming from MPA) dx=rand*(xBest(d)-x(n,d));
2)(MPA) if n>N/2 dx=progress2*randn*randn*(xBest(d)-x(n,d));
   else dx=rand*levyCoeff*levy(1,1,levyBeta,0)/strateg2Coeff*(xBest(d)-x(n,d));
3) (MPA) dx=progress2*levyCoeff*levy(1,1,levyBeta,0)^2*(xBest(d)-x(n,d));
4) (MPA) Back (restore previous position if better). BACK
5) (MPA) See aroundSearch . AROUND

6) (PSO) v(n,d)=w*v(n,d)+c*rand*(informBestX(p,d)-x(n,d)) ...
                        + c*rand*(xBest(d)-x(n,d));
                    x(n,d)=x(n,d)+v(n,d);

7)  (APS) Expansion
8)  (APS) Contraction
9)  Local search. LOCAL
10)  (APS) Partial restart (reinit)
11)  (APS) Stagnation detection STAGN
12) (APS) Selection (keep best agents). KEEP
13) (DE-like)  dx=coeff*rand*(x(randperm(N),:) -x(randperm(N),:)); 
14) (EDA)
   < see code below >
15) AOA (Archimedes Optimization Algorithm)
    < see code below >

%}
%==========================================================================
function [fBest,xBestReal,fBestEvolution,FEs]=MAMSO(run, N,FEmax,LB,UB,D,...
    fobj,quantis)
global allPos allPosSave
global verbose
global diamPlot
global fidResults

E=exp(1);

% ---------------------------------------------------------------Parameters
nbStrat=15; % Total number of strategies
% All parameters, so that any set of strategies can be used
% Of course, depending on this set, some parameters are useless
aroundCoeff=0.2;  %  Original MPA 0.2
aroundOption=0;
c = 0.5 + log(2); % PSO (SPSO 2007)
DEcoeff=0.5;
expAlpha=0.5; % To modify expansion
expBeta=0.75; %   For stagnation detection
K=2; % Mean number of informers (PSO)
keepRate =0.9 ; %  For elitist selection
levyBeta=1; % Original MPA: 1.5.
levyCoeff=0.05; % MPA

partN=0.5; % Part of population around the current best
           % to compute the diameter.

progressCoeff1=2;
progressCoeff2=progressCoeff1;
progressOption=0; % TO COMPLETE. Not used yet.
restartOption=0; % TO COMPLETE. Not used yet.

stagnOption=2; % TO COMPLETE. Not used yet.
strateg1Coeff=2;
strateg2Coeff=strateg1Coeff;

w=1/(2*log(2));  %  PSO (SPSO 2007)

C1=2;  % AOA-like
uAOA=0.9;

%***********                      Set of strategies to successively use
%sequenceList=[6*ones(1,210) 15*ones(1,15) 14*ones(1,15)]; 
%sequenceList=[6*ones(1,200) 3*ones(1,40) ];
sequenceList=[15];
%sequenceList=[6*ones(1,181) 15*ones(1,21) 14*ones(1,11)]; 
% In order (sequence). Can perfectly contain
% several times the same strategy
%outLoop=[ 4 5 13  ];  % on/off. Order is not important.
outLoop=[];
% Just once for each strategy
% Note: 4 is part of AOA (strategy 15 in sequencelList), 
sequenceOption=0; % 0 => For sequenceList.
%      Next strategy (cyclically) after each iteration
% 1 => according to FEs vs FEmax (needed for MPA, for example)
% 2 => according to stagnation detection (needs 11 in sequenceList or
%      outLoop)

localFrequency=0.1; % If outLoop contains 9,  this is the probability
                    % that local search is performed after the iteration
                    % A small value, like 0.1, seems to be a good choice
localOption=0; % 0 => Octave or Matlab 'native' search
               % 1 => A rudimentary pseudo-gradient method coded here 
               
aroundFrequency=1;

               
quantOnce=false; % Experimental. If true, and if discretisation is needed
% do it just once at the very end. Probably a bad idea.

%---------------------------------------------------------- Save parameters
if run==1
    fprintf(fidResults,'aroundCoeff %f \n',aroundCoeff);
    fprintf(fidResults,'aroundOption %i \n',aroundOption);
    fprintf(fidResults,'c %f \n',c);
    fprintf(fidResults,'DEcoeff %f \n',DEcoeff);
    fprintf(fidResults,'expAlpha %f \n',expAlpha);
    fprintf(fidResults,'expBeta %f \n',expBeta);
    fprintf(fidResults,'K %i \n',K);
    fprintf(fidResults,'keepRate %f \n',keepRate);
    fprintf(fidResults,'levyBeta %f \n',levyBeta);
    fprintf(fidResults,'levyCoeff %f \n',levyCoeff);
    fprintf(fidResults,'partN %f \n',partN);
    fprintf(fidResults,'progressCoeff1 %f \n',progressCoeff1);
    fprintf(fidResults,'strateg1Coeff %f \n',strateg1Coeff);
    fprintf(fidResults,'w %f \n',w);
    fprintf(fidResults,'C1 %f \n',C1);
    fprintf(fidResults,'uAOA %f \n',uAOA);
    fprintf(fidResults,'sequenceList ');
    for s=1:length(sequenceList)
        fprintf(fidResults,'%i ',sequenceList(s));
    end
    fprintf(fidResults,'\n');
    fprintf(fidResults,'outLoop ');
    for s=1:length(outLoop)
        fprintf(fidResults,'%i ',outLoop(s));
    end
    fprintf(fidResults,'\n');
    fprintf(fidResults,'sequenceOption %i \n',sequenceOption);
    fprintf(fidResults,'localFrequency %f \n',localFrequency);
    fprintf(fidResults,'localOption %i \n',localOption);
    if quantOnce str='true'; else str='false'; end
    fprintf(fidResults,'quantOnce %s \n',str);
    fprintf(fidResults,'Run FEs fBest');
end
%***************************************************************************
APS= sum(sequenceList==7)+sum(sequenceList==8)>0;
PSO= sum(sequenceList==6)>0;
AOA=sum(sequenceList==15)>0;

% These are triggered only after the loop on agents
BACK=sum(outLoop==4)>0;
AROUND=sum(outLoop==5)>0;
LOCAL=sum(outLoop==9)>0;
RESTART=sum(outLoop==10>0);
STAGN=sum(outLoop==11)>0;
KEEP=sum(outLoop==12)>0;
DE=sum(outLoop==13)>0;

if APS
    % As this is memory consuming, do not do it if not necessary
    triosAll=nchoosek(1:N,3);
end

%----------------------------------------------------------- Initialisation
[xReal,fitness,FEs]=initialisation(N,D,UB,LB,fobj,quantis,0,quantOnce);
[fBest,Ind]=min(fitness);
if allPosSave allPos=[allPos;xReal]; end

% Normalise
x=normX(xReal,LB,UB,0);
xBest=x(Ind,:);

diam0=partialDiameter(x,xBest,partN); % Around xBest
diam(1)=diam0;
diamFile=fopen('diameter','w');

if PSO
    v=zeros(N,D); % Initial velocity
    informers=inform(N,K); % Initial topology
    informBestX=x;
    informBestF=fitness;
end

if AOA
    den=rand(N,D);
    vol=rand(N,D);
    accNorm=rand(N,D);

    denBest=den(Ind,:);
    volBest=vol(Ind,:);
    accBest=accNorm(Ind,:);
end

strat=zeros(1,nbStrat); % Just to count how many time each strategy is used
iter=0; % The number of iterations is needed for some algorithms (not all)

%---------------------------------------------  Loop acording to the budget
s=1; % Rand in sequenceList of the first stragegy to use
keepSize=max(1,round(N*keepRate)); % If strategy KEEP is used

while FEs<=FEmax
    FEprev=FEs;
    fitPrev=fitness;
    xPrev=x;
    iter=iter+1;
    strateg=sequenceList(s);
    
    % -----------------------------------------------------Prepare strategy
    u=FEs/FEmax;
    prog=u-(N-1)/FEmax;
    phase=ceil(3*prog); % 1, 2, 3
    progress=(1-prog)^(progressCoeff1*prog);
    progress2=progress/progressCoeff2;
    
    mu=xBest; % For EDA. Just in case you want to use another mu
   sigm=0.5*diam(iter)*ones(1,D); % See strategy 14
    
    if AOA
        dAOA=E*(1-u)^2;
        acc=accNorm;
       
        for n=1:N
            mr=randi(N);
            raoa1=rand;
            raoa2=rand; 
            den(n,:)=den(n,:)+raoa1*(denBest-den(n,:));
            vol(n,:)=vol(n,:)+raoa1*(volBest-vol(n,:));
             if rand<1-2*u
                accTemp(n,:)=((den(mr,:).*vol(mr,:).*acc(mr,:)))...
                    ./(den(n,:).*vol(n,:))*raoa2;
            else
                accTemp(n,:)=(denBest.*volBest.*accBest)...
                    ./(den(n,:).*vol(n,:))*raoa2; 
            end
        end
        accNorm=((uAOA*(accTemp-min(accTemp(:))))./(max(accTemp(:))...
            -min(accTemp(:))))+1-uAOA;
    end
    
    if APS
        trios=triplet(N,N,triosAll);
    end
    
    % -------------------------------------------------------Apply strategy
    strat(strateg)=strat(strateg)+1; % Just to count
    for n=1:N
        if PSO
         % Finds the best informer
            informPSO= informBest(n,informers,informBestF);
        end
        for d=1:D
            switch strateg
                case 1  % Around the best
                    r=rand;
                    dx=r*(xBest(d)-x(n,d));
                    x(n,d)=x(n,d)+rand*dx/strateg1Coeff;
                    
                case 2
                    if n>N/2 %  For the first half of the population,
                        % larger moves around the best
                        rn=randn; % or randG(n,d)
                        dx=rn*rn*(xBest(d)-x(n,d));
                        %rn=bellLike(-0.5,0.5,1,5);
                        %dx=rn*(xBest(d)-x(n,d));
                        
                        x(n,d)=xBest(d)+progress2*dx;
                        
                    else  % For the second half,  even larger moves
                        % with sometimes 'outsiders'
                        rl=levyCoeff*levy(1,1,levyBeta,0); % or randL(n,d)
                        dx=rl*(xBest(d)-x(n,d));
                        x(n,d)=x(n,d)+rand*dx/strateg2Coeff;
                    end
                    
                case 3  % For some agents large moves (outsiders)
                    rl=levyCoeff*levy(1,1,levyBeta,0); % or randL(n,d)
                    dx=rl*rl*(xBest(d)-x(n,d));
                    x(n,d)=xBest(d)+progress2*dx;
                    
                case 6 % PSO like (synchronous mode)
                    v(n,d)=w*v(n,d)+c*rand*(informBestX(informPSO,d)-x(n,d)) ...
                        + c*rand*(xBest(d)-x(n,d));
                    x(n,d)=x(n,d)+v(n,d);
                    
                case 7 % Expansion (cf. APS)
                    pv=probaV(x);
                    r1=trios(n,1); r2=trios(n,2); r3=trios(n,3);
                    xBefore=x(n,:);
                    for d=1:D
                        if rand < pv % Semi-reflection
                            x(n,d)=x(r3,d)+...
                                (1+expAlpha)*((x(r1,d)+x(r2,d))/2-x(r3,d));
                        else
                            x(n,d)=xBefore(d);
                        end
                    end
                    
                    % It might happen that the position is not modified
                    % We could check it and not evaluate it, but as it
                    % happens rarely, we accept to 'waste' a few
                    % evaluations (two times the same position)
                    
                case 8 % Contraction (cf. APS)
                    pv=probaV(x);
                    r1=trios(n,1); r2=trios(n,2); r3=trios(n,3);
                    xBefore=x(n,:);
                    for d=1:D
                        if rand < pv
                            contract=2*rand; % To reduce bias. Note this is
                            % not always really 'contraction'
                            
                            cc=bellLike(0,1, 3,3); %Three random numbers
                            % from a bell-like distribution
                            cc=cc/sum(c); % => sum(c)=1
                            x(n,d)=(cc(1)*x(r1,d)+cc(2)*x(r2,d)+cc(3)*x(r3,d))*contract;
                            %{
                              Might be outside the search space, because
                              'contract'. On the other hand, with the true
                              weighted gravity center,results are not
                              as good.
                            %}
                        else
                            x(n,d)=xBefore(d);
                        end % if rand < pv
                    end % for d=1:D
                    
                    % It might happen that the position is not modified
                    % ...
                    
                case 14 % EDA
                   %dxb=abs(mu(d)-x(n,d));
                   dxb=sigm(d);
                   a=mu(d)-dxb; b=mu(d)+dxb;
                    x(n,d)=bellLike(a,b,1,5);
                    
                    % x(n,d)=mu(d)+randn*sigm(d); % Another way
                    
                case 15 % AOA
                   if rand<1-u
                        mrand=randi(N);
                        x(n,d)=xPrev(n,d)+C1*rand*accNorm(n,d).*...
                            (xPrev(mrand,d)-xPrev(n,d))*dAOA;
                    else
                        x(n,d)=xBest(d)+3*C1*rand*accNorm(n,d).*...
                            (xBest(d)-xPrev(n,d))*dAOA;
                   end
                  
                    
            end % switch strategy
        end   % for d=1:D
    end   % for n=1:N
    
    %-------------------- After the loop on agents
    if PSO && strateg~=6  % If PSO is used, but not this time ...
        v=x-xPrev; % ... velocity has to be updated
    end
    
    BACK=BACK || strateg==4;
    AROUND=AROUND || strateg==5;
    LOCAL=LOCAL || strateg==9;
    RESTART=RESTART || strateg==10;
    STAGN=STAGN || strateg==11;
    KEEP=KEEP || strateg==12;
    DE=DE || strateg==13;
    
    % Just for information
    if BACK strat(4)=strat(4)+1; end
    if AROUND strat(5)=strat(5)+1; end
    if LOCAL strat(9)=strat(9)+1; end
    if RESTART strat(10)=strat(10)+1; end
    if STAGN strat(11)=strat(11)+1; end
    if KEEP strat(12)=strat(12)+1; end
    if DE strat(13)=strat(13)+1; end
 
    stagn=false;
    
    if AROUND && rand<aroundFrequency
        r=rand;
        progress=diam(iter);
        if r<aroundCoeff % Random move around
            x=aroundSearch(x,aroundCoeff,progress);
        end
    end
    
    if DE
        r=rand;
        if r>aroundCoeff
            x=DElike(x,DEcoeff);
        end
    end
    
    % Confinement
    x=confin(x);
    
    % Back to real search space and evaluation
    xReal=normX(x,LB,UB,1);
    if ~quantOnce
        xReal=quantify(xReal,quantis);
    end
    [fitness,FEs]=fEvalAll(xReal,fobj,FEs);
    
    if BACK % ... back if deterioration
        [fitness,x]=updateXF(fitPrev,xPrev,fitness,x);
    end
    
    if LOCAL && rand<=localFrequency
        if BACK
            xReal=normX(x,LB,UB,1); % I know, not very clever to sometimes
            % do that twice
        end
        [xReal,fitness,FEs]=...
            localSearch(xReal,fitness,FEs,fobj,LB,UB,FEmax,quantis,...
            quantOnce,diam(iter), localOption);
        
        % Re-normalise
        if ~quantOnce
            xReal=quantify(xReal,quantis);
        end
      
        x=normX(xReal,LB,UB,0);
        x=confin(x);
    end
    
    
    if KEEP % Keep the best ones
        [fKeep,Ind]=sort(fitPrev);
        xKeep=xPrev(Ind,:);
        
        xAll=[xKeep(1:keepSize,:);x];
        fAll=[fKeep(1:keepSize) fitness];
        [fSort,Ind]=sort(fAll);
        xSort=xAll(Ind,:);

        x=xSort(1:N,:);
        f=fSort(1:N);
    end
    
    if STAGN
        % Detect stagnation
        stagn=stagnation(fitPrev,fitness, FEprev, FEs,expBeta) ;
    end
    
    if PSO % Update the memory swarm(synchronous mode)
        for n=1:N
            if informBestF(n)>fitness(n)
                informBestF(n)=fitness(n);
                informBestX(n,:)=x(n,:);
            end
        end
    end
    %{
    if PSO % Update the memory swarm(synchronous mode)
          % Variant with more randomness
        neigh=randperm(N);
        for n=1:N
            m=neigh(n);
            if informBestF(m)>=fitness(m)
                informBestF(m)=fitness(m);
                informBestX(m,:)=x(m,:);
            end
        end
    end
    %}
    
    
    if RESTART
        if STAGN reinit=stagn; else reinit=true; end
        if reinit
            [fitness,Ind]=sort(fitness);
            x=x(Ind,:);
            [xInit,fInit,FEs]=initialisation(N-keepSize,D,LB,UB, fobj,...
                quantis,FEs,quantOnce);
            xInit=normX(xInit,LB,UB,0); % Normalise, because initialisation
                                        % is in the real search space
            x=[x(1:keepSize,:) ; xInit];
            fitness=[fitness(1:keepSize) fInit];
        end
    end
    
    %--------------------
    [fBest,xBest,Ind]=bestXF(x,fitness,fBest,xBest);
    fBestEvolution(iter)=fBest;
    
    if allPosSave allPos=[allPos;normX(x,LB,UB,1)]; end

    
    if AOA
        denBest=den(Ind,:);
        volBest=vol(Ind,:);
        accBest=accNorm(Ind,:);
    end
    
    diam(iter+1)=partialDiameter(x,xBest,partN);
    
    % Select strategy
    switch sequenceOption
        case 0
            s=s+1;
            
        case 1
            prog=(FEs-N+1)/FEmax;
            phaseNew=ceil(3*prog);
            % This is supposed to work with a sequenceList that contains
            % [ 1 2 3]
            if phaseNew>phase s=s+1; end
            
        case 2
            if stagn s=s+1; end
            
    end
    if s>length(sequenceList) s=1; end
    
    
end % While FEs<FEmax

if verbose
    fprintf('\n Strategies \n'); disp(strat); % Just for information
    fprintf('\n  %i iterations',iter);
end

% Real best position
xBestReal=normX(xBest,LB,UB,1);

if quantOnce % Quantify only at the very end
    xBestReal=quantify(xBestReal,quantis);
    [fBest,FEs]=fEvalAll(xBestReal,fobj,FEs); % Actually just one evaluation
end

if diamPlot % WARNING: one plot by run
    %  Evolution of the nPart*N population diameter
    % (around the best at each iteration)
    figure
    FE=1:length(diam/diam0);
    plot(FE,diam);
    % set(gca,'xtick',[]);
    % set(gca,'ytick',[]);
    xlabel('Iteration');
    ylabel('Relative diameter');
    title('Partial-population diameter')
end
end
